# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import gedit

from gtk import ActionGroup, STOCK_CONVERT, CheckMenuItem, MenuItem, Image
from gobject import timeout_add
from os import getcwd

from DocumentWrapper import DocumentWrapper
from StructurePanel import StructurePanel
from SymbolsPanel import SymbolsPanel
from SnippetsPanel import SnippetsPanel
from MainToolbar import MainToolbar
from Builder import Builder
from Settings import Settings
from common import PLUGIN_PATH
from AutoCompletePopup import AutoCompletePopup



DBUS_BUS_NAME = 'org.gedit.LaTeXPlugin'
DBUS_OBJECT_PATH = '/org/gedit/DBUSInterface'

class LaTeXPluginInstance:
	def __init__(self, plugin, window):
		self.plugin, self.window = plugin, window
		self.document_wrappers = {}			## maps GeditView to DocumentWrapper objects 
		self.active_wrapper = None
		self.current_context_id = None
		self.settings = Settings()
		self.builder = Builder(self)
		self.__insert_widgets()
		self.save_action = self.ui_manager.get_action("/MenuBar/FileMenu/FileSaveMenu")
		self.auto_complete_popup = AutoCompletePopup()
		
		if self.settings["dvi-search-enabled"]:
			self.__load_dbus_interface()
		
	def __load_dbus_interface(self):
		try:
			import dbus.service
			if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
				import dbus.glib
			
			class DBUSInterface (dbus.service.Object):
				"""
				The DBUSInterface is listening for DVI inverse search commands and
				scrolls the document to the appropriate line
				"""
				def __init__(self, instance, bus_name, object_path=DBUS_OBJECT_PATH):
					dbus.service.Object.__init__(self, bus_name, object_path)
					self.instance = instance

				# This is for dbus inverse search
				@dbus.service.method('org.gedit.LaTeXPluginIFace')
				def inverse(self, filename, line):
					print "Got inverse DVI search: " + filename + " " + str(line)
					
					self.instance.set_active_tab_by_filename(filename)
					wrapper = self.instance.get_active_wrapper()
					wrapper.jump_to_line(line)


				# adding this makes things work---don't know why!
				@dbus.service.method('org.gedit.LaTeXPluginIFace')
				def dummy(self):
					pass
		
			# DBUS interface
			bus_name = dbus.service.BusName(DBUS_BUS_NAME, bus=dbus.SessionBus())
			self.dbus_interface = DBUSInterface(self, bus_name)
			
		except ImportError:
			print "Warning: failed to import DBUS bindings"
	
	def update(self):
		active_view = self.window.get_active_view()
		if not active_view:
			return
			
		try:
			## activate the DocumentWrapper connected to that GeditView
			wrapper = self.document_wrappers[active_view]
			wrapper.activate()
		except KeyError:
			## create a new DocumentWrapper
			wrapper = DocumentWrapper(self, active_view)
			self.document_wrappers[active_view] = wrapper
		
		if wrapper != self.active_wrapper:
			## the active wrapper has changed...
			self.structure_panel.set_structure_model(wrapper.get_structure_model())	## display its structure...
			self.structure_panel.set_labels_model(wrapper.get_labels_model())		## ...and labels/bibitems
		
			self.active_wrapper = wrapper
		
	def deactivate(self):
		self.__remove_widgets()
	
	def __pop_status_message(self):
		status_bar = self.window.get_statusbar()
		if self.current_context_id:
			status_bar.pop(self.current_context_id)
		self.current_context_id = None
		
		return False		## do not call again
	
	def push_status_message(self, message):
		"""
		Push a message to the window's status bar and start a timer that pops it again
		"""
		status_bar = self.window.get_statusbar()
		self.current_context_id = status_bar.get_context_id(message)
		status_bar.push(self.current_context_id, message)
	
		timeout_add(3000, self.__pop_status_message)
		
	def get_active_wrapper(self):
		return self.active_wrapper
	
	def save_active_document(self):
		self.save_action.activate()
	
	def set_active_tab_by_filename(self, filename):
		"""
		Activate the document tab with a specific filename
		"""
		if filename[0] != "/":
			filename = "%s/%s" % (getcwd(), filename)	## rubber returns relative filenames
		
		## check if the tab is already active
		if self.active_wrapper.get_filename() == filename:
			return
		
		## find the tab and activate it
		for wrapper in self.document_wrappers.values():
			if wrapper.get_filename() == filename:
				## TODO: document_wrappers must be kept clean to avoid errors here!
				tab = gedit.tab_get_from_document(wrapper.get_document())
				self.window.set_active_tab(tab)
				return
		
		## try to open the document in gedit
		print self.window.create_tab_from_uri(filename, None, False, True)
		
		print "set_active_tab_by_filename(): %s, document not found" % filename
	
	def __insert_widgets(self):
		## attach panels
		self.structure_panel = StructurePanel(self)
		self.symbols_panel = SymbolsPanel(self)
		self.snippets_panel = SnippetsPanel(self)
		
		img_outline = Image()
		img_outline.set_from_file(PLUGIN_PATH + "/pixmaps/outline.png")
		
		img_symbols = Image()
		img_symbols.set_from_file(PLUGIN_PATH + "/pixmaps/symbols.png")
		
		img_snippets = Image()
		img_snippets.set_from_file(PLUGIN_PATH + "/pixmaps/snippet.png")
		
		self.window.get_side_panel().add_item(self.structure_panel, _("Document Structure"), img_outline)
		self.window.get_side_panel().add_item(self.symbols_panel, _("Symbol Map"), img_symbols)
		self.window.get_side_panel().add_item(self.snippets_panel, _("Snippets"), img_snippets)
		
		## insert menu items
		self.ui_manager = self.window.get_ui_manager()
		self.action_group = ActionGroup("LaTeXPluginActions")
		self.action_group.add_actions([
				("Comment", None, _("Comment"), "<Ctrl><Alt>C", None, lambda x: self.get_active_wrapper().comment_selection()),
				("Uncomment", None, _("Uncomment"), "<Ctrl><Alt>U", None, lambda x: self.get_active_wrapper().comment_selection(uncomment=True)),
				
				("Build", STOCK_CONVERT, _("Build LaTeX Document"), None, None, None),
				("BuildPDF", None, "PDF", "F5", _("Builds a PDF file from the document"), lambda x: self.builder.build_pdf()),
				("BuildDVI", None, "DVI", "<Ctrl>F5", _("Builds a DVI file from the document"), lambda x: self.builder.build_dvi()),
				("BuildPS", None, "PostScript", "<Shift><Ctrl>F5", _("Builds a PostScript file from the document"), lambda x: self.builder.build_ps()),
				("BuildImage", None, _("Image") + "...", None, _("Builds an image from the document"), lambda x: self.builder.build_image())
		])
		
		self.ui_manager.insert_action_group(self.action_group, -1)
		self.ui_id = self.ui_manager.add_ui_from_file(PLUGIN_PATH + "/ui.xml")
		
		## append custom profiles
		build_menu = self.ui_manager.get_widget("/MenuBar/ToolsMenu/ToolsOps_2/BuildMenu").get_submenu()
		
		for profile in self.settings.get_profiles():
			item = MenuItem(profile.name)
			item.connect("activate", lambda x: self.builder.build_custom(profile.key))
			item.show()
			build_menu.append(item)
		
		## hook the toolbar
		self.toolbar = MainToolbar(self)
		self.main_box = self.window.get_children()[0]
		self.main_box.pack_start(self.toolbar, False)
		self.main_box.reorder_child(self.toolbar, 2)
		
		## add a CheckMenuItem for the toolbar
		self.item_show_toolbar = CheckMenuItem(_("LaTeX Toolbar"))
		self.item_show_toolbar.set_active(True)
		self.ui_manager.get_widget("/MenuBar/ViewMenu").get_submenu().insert(self.item_show_toolbar, 2)
		self.item_show_toolbar.show()
		self.item_show_toolbar.connect("toggled", self.__on_toolbar_toggled)
	
	def __remove_widgets(self):
		## remove panels
		self.window.get_side_panel().remove_item(self.symbols_panel)
		self.window.get_side_panel().remove_item(self.structure_panel)
		
		## remove menu items
		self.ui_manager.remove_ui(self.ui_id)
		self.ui_manager.remove_action_group(self.action_group)
		self.ui_manager.ensure_update()
		
		## remove toolbar
		self.main_box.remove(self.toolbar)
		
		## destroy
		self.toolbar, self.symbols_panel, self.structure_panel = None, None, None
	
	def __on_toolbar_toggled(self, checkmenuitem):
		if checkmenuitem.get_active():
			self.toolbar.show()
		else:
			self.toolbar.hide()
	
class LaTeXPlugin(gedit.Plugin):
	def __init__(self):
		gedit.Plugin.__init__(self)
		self.TAG = "LaTeXPluginInstance"
		
	def __getInstance(self, window):
		return window.get_data(self.TAG)
	
	def __setInstance(self, window, instance):
		window.set_data(self.TAG, instance)
	
	def activate(self, window):
		"""Called when a new gedit window is created"""
		self.__setInstance(window, LaTeXPluginInstance(self, window))
	
	def deactivate(self, window):
		"""Called when a window is destroyed"""
		self.__getInstance(window).deactivate()
		self.__setInstance(window, None)
	
	def update_ui(self, window):
		"""Called when the window is updated (tab change, menu update etc)"""
		self.__getInstance(window).update()
	
	def create_configure_dialog(self):
		"""Called when the plugin is to be configured (must return a gtk.Window)"""
		from ConfigureDialog import ConfigureDialog
		return ConfigureDialog()
		
